from flask import json

from authlib.jose import jwt
from authlib.oauth2.rfc7591 import ClientMetadataClaims as OAuth2ClientMetadataClaims
from authlib.oauth2.rfc7591 import (
    ClientRegistrationEndpoint as _ClientRegistrationEndpoint,
)
from authlib.oidc.registration import ClientMetadataClaims as OIDCClientMetadataClaims
from tests.util import read_file_path

from .models import Client
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class ClientRegistrationEndpoint(_ClientRegistrationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def authenticate_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            request.user_id = 1
            return auth_header

    def resolve_public_key(self, request):
        return read_file_path("rsa_public.pem")

    def save_client(self, client_info, client_metadata, request):
        client = Client(user_id=request.user_id, **client_info)
        client.set_client_metadata(client_metadata)
        db.session.add(client)
        db.session.commit()
        return client


class OAuthClientRegistrationTest(TestCase):
    def prepare_data(self, endpoint_cls=None, metadata=None):
        app = self.app
        server = create_authorization_server(app)

        if endpoint_cls:
            server.register_endpoint(endpoint_cls)
        else:

            class MyClientRegistration(ClientRegistrationEndpoint):
                def get_server_metadata(self):
                    return metadata

            server.register_endpoint(MyClientRegistration)

        @app.route("/create_client", methods=["POST"])
        def create_client():
            return server.create_endpoint_response("client_registration")

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

    def test_access_denied(self):
        self.prepare_data()
        rv = self.client.post("/create_client", json={})
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "access_denied")

    def test_invalid_request(self):
        self.prepare_data()
        headers = {"Authorization": "bearer abc"}
        rv = self.client.post("/create_client", json={}, headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

    def test_create_client(self):
        self.prepare_data()
        headers = {"Authorization": "bearer abc"}
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

    def test_software_statement(self):
        payload = {"software_id": "uuid-123", "client_name": "Authlib"}
        s = jwt.encode({"alg": "RS256"}, payload, read_file_path("rsa_private.pem"))
        body = {
            "software_statement": s.decode("utf-8"),
        }

        self.prepare_data()
        headers = {"Authorization": "bearer abc"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

    def test_no_public_key(self):
        class ClientRegistrationEndpoint2(ClientRegistrationEndpoint):
            def get_server_metadata(self):
                return None

            def resolve_public_key(self, request):
                return None

        payload = {"software_id": "uuid-123", "client_name": "Authlib"}
        s = jwt.encode({"alg": "RS256"}, payload, read_file_path("rsa_private.pem"))
        body = {
            "software_statement": s.decode("utf-8"),
        }

        self.prepare_data(ClientRegistrationEndpoint2)
        headers = {"Authorization": "bearer abc"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "unapproved_software_statement")

    def test_scopes_supported(self):
        metadata = {"scopes_supported": ["profile", "email"]}
        self.prepare_data(metadata=metadata)

        headers = {"Authorization": "bearer abc"}
        body = {"scope": "profile email", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        body = {"scope": "profile email address", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_response_types_supported(self):
        metadata = {"response_types_supported": ["code"]}
        self.prepare_data(metadata=metadata)

        headers = {"Authorization": "bearer abc"}
        body = {"response_types": ["code"], "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        # https://www.rfc-editor.org/rfc/rfc7591.html#section-2
        # If omitted, the default is that the client will use only the "code"
        # response type.
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        body = {"response_types": ["code", "token"], "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_grant_types_supported(self):
        metadata = {"grant_types_supported": ["authorization_code", "password"]}
        self.prepare_data(metadata=metadata)

        headers = {"Authorization": "bearer abc"}
        body = {"grant_types": ["password"], "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        # https://www.rfc-editor.org/rfc/rfc7591.html#section-2
        # If omitted, the default behavior is that the client will use only
        # the "authorization_code" Grant Type.
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        body = {"grant_types": ["client_credentials"], "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_token_endpoint_auth_methods_supported(self):
        metadata = {"token_endpoint_auth_methods_supported": ["client_secret_basic"]}
        self.prepare_data(metadata=metadata)

        headers = {"Authorization": "bearer abc"}
        body = {
            "token_endpoint_auth_method": "client_secret_basic",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        body = {"token_endpoint_auth_method": "none", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")


class OIDCClientRegistrationTest(TestCase):
    def prepare_data(self, metadata=None):
        self.headers = {"Authorization": "bearer abc"}
        app = self.app
        server = create_authorization_server(app)

        class MyClientRegistration(ClientRegistrationEndpoint):
            def get_server_metadata(self):
                return metadata

        server.register_endpoint(
            MyClientRegistration(
                claims_classes=[OAuth2ClientMetadataClaims, OIDCClientMetadataClaims]
            )
        )

        @app.route("/create_client", methods=["POST"])
        def create_client():
            return server.create_endpoint_response("client_registration")

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

    def test_application_type(self):
        self.prepare_data()

        # Nominal case
        body = {
            "application_type": "web",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["application_type"], "web")

        # Default case
        # The default, if omitted, is that any algorithm supported by the OP and the RP MAY be used.
        body = {
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["application_type"], "web")

        # Error case
        body = {
            "application_type": "invalid",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_token_endpoint_auth_signing_alg_supported(self):
        metadata = {
            "token_endpoint_auth_signing_alg_values_supported": ["RS256", "ES256"]
        }
        self.prepare_data(metadata)

        # Nominal case
        body = {
            "token_endpoint_auth_signing_alg": "ES256",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["token_endpoint_auth_signing_alg"], "ES256")

        # Default case
        # The default, if omitted, is that any algorithm supported by the OP and the RP MAY be used.
        body = {
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")

        # Error case
        body = {
            "token_endpoint_auth_signing_alg": "RS512",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_subject_types_supported(self):
        metadata = {"subject_types_supported": ["public", "pairwise"]}
        self.prepare_data(metadata)

        # Nominal case
        body = {"subject_type": "public", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["subject_type"], "public")

        # Error case
        body = {"subject_type": "invalid", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_id_token_signing_alg_values_supported(self):
        metadata = {"id_token_signing_alg_values_supported": ["RS256", "ES256"]}
        self.prepare_data(metadata)

        # Default
        # The default, if omitted, is RS256.
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["id_token_signed_response_alg"], "RS256")

        # Nominal case
        body = {"id_token_signed_response_alg": "ES256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["id_token_signed_response_alg"], "ES256")

        # Error case
        body = {"id_token_signed_response_alg": "RS512", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_id_token_encryption_alg_values_supported(self):
        metadata = {"id_token_encryption_alg_values_supported": ["RS256", "ES256"]}
        self.prepare_data(metadata)

        # Default case
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertNotIn("id_token_encrypted_response_enc", resp)

        # If id_token_encrypted_response_alg is specified, the default
        # id_token_encrypted_response_enc value is A128CBC-HS256.
        body = {"id_token_encrypted_response_alg": "RS256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["id_token_encrypted_response_enc"], "A128CBC-HS256")

        # Nominal case
        body = {"id_token_encrypted_response_alg": "ES256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["id_token_encrypted_response_alg"], "ES256")

        # Error case
        body = {"id_token_encrypted_response_alg": "RS512", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_id_token_encryption_enc_values_supported(self):
        metadata = {
            "id_token_encryption_enc_values_supported": ["A128CBC-HS256", "A256GCM"]
        }
        self.prepare_data(metadata)

        # Nominal case
        body = {
            "id_token_encrypted_response_alg": "RS256",
            "id_token_encrypted_response_enc": "A256GCM",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["id_token_encrypted_response_alg"], "RS256")
        self.assertEqual(resp["id_token_encrypted_response_enc"], "A256GCM")

        # Error case: missing id_token_encrypted_response_alg
        body = {"id_token_encrypted_response_enc": "A256GCM", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

        # Error case: alg not in server metadata
        body = {"id_token_encrypted_response_enc": "A128GCM", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_userinfo_signing_alg_values_supported(self):
        metadata = {"userinfo_signing_alg_values_supported": ["RS256", "ES256"]}
        self.prepare_data(metadata)

        # Nominal case
        body = {"userinfo_signed_response_alg": "ES256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["userinfo_signed_response_alg"], "ES256")

        # Error case
        body = {"userinfo_signed_response_alg": "RS512", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_userinfo_encryption_alg_values_supported(self):
        metadata = {"userinfo_encryption_alg_values_supported": ["RS256", "ES256"]}
        self.prepare_data(metadata)

        # Nominal case
        body = {"userinfo_encrypted_response_alg": "ES256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["userinfo_encrypted_response_alg"], "ES256")

        # Error case
        body = {"userinfo_encrypted_response_alg": "RS512", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_userinfo_encryption_enc_values_supported(self):
        metadata = {
            "userinfo_encryption_enc_values_supported": ["A128CBC-HS256", "A256GCM"]
        }
        self.prepare_data(metadata)

        # Default case
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertNotIn("userinfo_encrypted_response_enc", resp)

        # If userinfo_encrypted_response_alg is specified, the default
        # userinfo_encrypted_response_enc value is A128CBC-HS256.
        body = {"userinfo_encrypted_response_alg": "RS256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["userinfo_encrypted_response_enc"], "A128CBC-HS256")

        # Nominal case
        body = {
            "userinfo_encrypted_response_alg": "RS256",
            "userinfo_encrypted_response_enc": "A256GCM",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["userinfo_encrypted_response_alg"], "RS256")
        self.assertEqual(resp["userinfo_encrypted_response_enc"], "A256GCM")

        # Error case: no userinfo_encrypted_response_alg
        body = {"userinfo_encrypted_response_enc": "A256GCM", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

        # Error case: alg not in server metadata
        body = {"userinfo_encrypted_response_enc": "A128GCM", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_acr_values_supported(self):
        metadata = {
            "acr_values_supported": [
                "urn:mace:incommon:iap:silver",
                "urn:mace:incommon:iap:bronze",
            ],
        }
        self.prepare_data(metadata)

        # Nominal case
        body = {
            "default_acr_values": ["urn:mace:incommon:iap:silver"],
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["default_acr_values"], ["urn:mace:incommon:iap:silver"])

        # Error case
        body = {
            "default_acr_values": [
                "urn:mace:incommon:iap:silver",
                "urn:mace:incommon:iap:gold",
            ],
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_request_object_signing_alg_values_supported(self):
        metadata = {"request_object_signing_alg_values_supported": ["RS256", "ES256"]}
        self.prepare_data(metadata)

        # Nominal case
        body = {"request_object_signing_alg": "ES256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["request_object_signing_alg"], "ES256")

        # Error case
        body = {"request_object_signing_alg": "RS512", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_request_object_encryption_alg_values_supported(self):
        metadata = {
            "request_object_encryption_alg_values_supported": ["RS256", "ES256"]
        }
        self.prepare_data(metadata)

        # Nominal case
        body = {
            "request_object_encryption_alg": "ES256",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["request_object_encryption_alg"], "ES256")

        # Error case
        body = {
            "request_object_encryption_alg": "RS512",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_request_object_encryption_enc_values_supported(self):
        metadata = {
            "request_object_encryption_enc_values_supported": [
                "A128CBC-HS256",
                "A256GCM",
            ]
        }
        self.prepare_data(metadata)

        # Default case
        body = {"client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertNotIn("request_object_encryption_enc", resp)

        # If request_object_encryption_alg is specified, the default
        # request_object_encryption_enc value is A128CBC-HS256.
        body = {"request_object_encryption_alg": "RS256", "client_name": "Authlib"}
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["request_object_encryption_enc"], "A128CBC-HS256")

        # Nominal case
        body = {
            "request_object_encryption_alg": "RS256",
            "request_object_encryption_enc": "A256GCM",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["request_object_encryption_alg"], "RS256")
        self.assertEqual(resp["request_object_encryption_enc"], "A256GCM")

        # Error case: missing request_object_encryption_alg
        body = {
            "request_object_encryption_enc": "A256GCM",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

        # Error case: alg not in server metadata
        body = {
            "request_object_encryption_enc": "A128GCM",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")

    def test_require_auth_time(self):
        self.prepare_data()

        # Default case
        body = {
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["require_auth_time"], False)

        # Nominal case
        body = {
            "require_auth_time": True,
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn("client_id", resp)
        self.assertEqual(resp["client_name"], "Authlib")
        self.assertEqual(resp["require_auth_time"], True)

        # Error case
        body = {
            "require_auth_time": "invalid",
            "client_name": "Authlib",
        }
        rv = self.client.post("/create_client", json=body, headers=self.headers)
        resp = json.loads(rv.data)
        self.assertIn(resp["error"], "invalid_client_metadata")



import json
import time

from authlib.oauth2.rfc6749.grants import RefreshTokenGrant as _RefreshTokenGrant

from .models import Client
from .models import OAuth2Token
from .models import User
from .oauth2_server import TestCase


class RefreshTokenGrant(_RefreshTokenGrant):
    def authenticate_refresh_token(self, refresh_token):
        try:
            item = OAuth2Token.objects.get(refresh_token=refresh_token)
            if item.is_refresh_token_active():
                return item
        except OAuth2Token.DoesNotExist:
            return None

    def authenticate_user(self, credential):
        return credential.user

    def revoke_old_credential(self, credential):
        now = int(time.time())
        credential.access_token_revoked_at = now
        credential.refresh_token_revoked_at = now
        credential.save()
        return credential


class RefreshTokenTest(TestCase):
    def create_server(self):
        server = super().create_server()
        server.register_grant(RefreshTokenGrant)
        return server

    def prepare_client(self, grant_type="refresh_token", scope=""):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            scope=scope,
            grant_type=grant_type,
            token_endpoint_auth_method="client_secret_basic",
            default_redirect_uri="https://a.b",
        )
        client.save()

    def prepare_token(self, scope="profile", user_id=1):
        token = OAuth2Token(
            user_id=user_id,
            client_id="client",
            token_type="bearer",
            access_token="a1",
            refresh_token="r1",
            scope=scope,
            expires_in=3600,
        )
        token.save()

    def test_invalid_client(self):
        server = self.create_server()
        self.prepare_client()
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "refresh_token", "refresh_token": "foo"},
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "refresh_token", "refresh_token": "foo"},
            HTTP_AUTHORIZATION=self.create_basic_auth("invalid", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_client")

    def test_invalid_refresh_token(self):
        self.prepare_client()
        server = self.create_server()
        auth_header = self.create_basic_auth("client", "secret")
        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "refresh_token"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_request")
        self.assertIn("Missing", data["error_description"])

        request = self.factory.post(
            "/oauth/token",
            data={"grant_type": "refresh_token", "refresh_token": "invalid"},
            HTTP_AUTHORIZATION=auth_header,
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_grant")

    def test_invalid_scope(self):
        server = self.create_server()
        server.scopes_supported = ["profile"]
        self.prepare_client()
        self.prepare_token()
        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "invalid",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_scope")

    def test_authorize_tno_scope(self):
        server = self.create_server()
        self.prepare_client()
        self.prepare_token()

        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertIn("access_token", data)

    def test_authorize_token_scope(self):
        server = self.create_server()
        self.prepare_client()
        self.prepare_token()

        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertIn("access_token", data)

    def test_revoke_old_token(self):
        server = self.create_server()
        self.prepare_client()
        self.prepare_token()

        request = self.factory.post(
            "/oauth/token",
            data={
                "grant_type": "refresh_token",
                "refresh_token": "r1",
                "scope": "profile",
            },
            HTTP_AUTHORIZATION=self.create_basic_auth("client", "secret"),
        )
        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertIn("access_token", data)

        resp = server.create_token_response(request)
        self.assertEqual(resp.status_code, 400)



from flask import json

from authlib.common.security import generate_token
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import grants
from authlib.oauth2.rfc7636 import CodeChallenge as _CodeChallenge
from authlib.oauth2.rfc7636 import create_s256_code_challenge

from .models import Client
from .models import CodeGrantMixin
from .models import User
from .models import db
from .models import save_authorization_code
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server


class AuthorizationCodeGrant(CodeGrantMixin, grants.AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class CodeChallenge(_CodeChallenge):
    SUPPORTED_CODE_CHALLENGE_METHOD = ["plain", "S256", "S128"]


class CodeChallengeTest(TestCase):
    def prepare_data(self, token_endpoint_auth_method="none"):
        server = create_authorization_server(self.app)
        server.register_grant(AuthorizationCodeGrant, [CodeChallenge(required=True)])

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()

        client_secret = ""
        if token_endpoint_auth_method != "none":
            client_secret = "code-secret"

        client = Client(
            user_id=user.id,
            client_id="code-client",
            client_secret=client_secret,
        )
        client.set_client_metadata(
            {
                "redirect_uris": ["https://a.b"],
                "scope": "profile address",
                "token_endpoint_auth_method": token_endpoint_auth_method,
                "response_types": ["code"],
                "grant_types": ["authorization_code"],
            }
        )
        self.authorize_url = "/oauth/authorize?response_type=code&client_id=code-client"
        db.session.add(client)
        db.session.commit()

    def test_missing_code_challenge(self):
        self.prepare_data()
        rv = self.client.get(self.authorize_url + "&code_challenge_method=plain")
        self.assertIn(b"Missing", rv.data)

    def test_has_code_challenge(self):
        self.prepare_data()
        rv = self.client.get(
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        self.assertEqual(rv.data, b"ok")

    def test_invalid_code_challenge(self):
        self.prepare_data()
        rv = self.client.get(
            self.authorize_url + "&code_challenge=abc&code_challenge_method=plain"
        )
        self.assertIn(b"Invalid", rv.data)

    def test_invalid_code_challenge_method(self):
        self.prepare_data()
        suffix = "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s&code_challenge_method=invalid"
        rv = self.client.get(self.authorize_url + suffix)
        self.assertIn(b"Unsupported", rv.data)

    def test_supported_code_challenge_method(self):
        self.prepare_data()
        suffix = "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s&code_challenge_method=plain"
        rv = self.client.get(self.authorize_url + suffix)
        self.assertEqual(rv.data, b"ok")

    def test_trusted_client_without_code_challenge(self):
        self.prepare_data("client_secret_basic")
        rv = self.client.get(self.authorize_url)
        self.assertEqual(rv.data, b"ok")

        rv = self.client.post(self.authorize_url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))

        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_missing_code_verifier(self):
        self.prepare_data()
        url = (
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("Missing", resp["error_description"])

    def test_trusted_client_missing_code_verifier(self):
        self.prepare_data("client_secret_basic")
        url = (
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        headers = self.create_basic_header("code-client", "code-secret")
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertIn("Missing", resp["error_description"])

    def test_plain_code_challenge_invalid(self):
        self.prepare_data()
        url = (
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": "bar",
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("Invalid", resp["error_description"])

    def test_plain_code_challenge_failed(self):
        self.prepare_data()
        url = (
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": generate_token(48),
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("failed", resp["error_description"])

    def test_plain_code_challenge_success(self):
        self.prepare_data()
        code_verifier = generate_token(48)
        url = self.authorize_url + "&code_challenge=" + code_verifier
        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": code_verifier,
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_s256_code_challenge_success(self):
        self.prepare_data()
        code_verifier = generate_token(48)
        code_challenge = create_s256_code_challenge(code_verifier)
        url = self.authorize_url + "&code_challenge=" + code_challenge
        url += "&code_challenge_method=S256"

        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        rv = self.client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": code_verifier,
                "client_id": "code-client",
            },
        )
        resp = json.loads(rv.data)
        self.assertIn("access_token", resp)

    def test_not_implemented_code_challenge_method(self):
        self.prepare_data()
        url = (
            self.authorize_url
            + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
        )
        url += "&code_challenge_method=S128"

        rv = self.client.post(url, data={"user_id": "1"})
        self.assertIn("code=", rv.location)

        params = dict(url_decode(urlparse.urlparse(rv.location).query))
        code = params["code"]
        self.assertRaises(
            RuntimeError,
            self.client.post,
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": generate_token(48),
                "client_id": "code-client",
            },
        )



import unittest

from authlib.oidc.discovery import OpenIDProviderMetadata
from authlib.oidc.discovery import get_well_known_url

WELL_KNOWN_URL = "/.well-known/openid-configuration"


class WellKnownTest(unittest.TestCase):
    def test_no_suffix_issuer(self):
        self.assertEqual(get_well_known_url("https://authlib.org"), WELL_KNOWN_URL)
        self.assertEqual(get_well_known_url("https://authlib.org/"), WELL_KNOWN_URL)

    def test_with_suffix_issuer(self):
        self.assertEqual(
            get_well_known_url("https://authlib.org/issuer1"),
            "/issuer1" + WELL_KNOWN_URL,
        )
        self.assertEqual(
            get_well_known_url("https://authlib.org/a/b/c"), "/a/b/c" + WELL_KNOWN_URL
        )

    def test_with_external(self):
        self.assertEqual(
            get_well_known_url("https://authlib.org", external=True),
            "https://authlib.org" + WELL_KNOWN_URL,
        )


class OpenIDProviderMetadataTest(unittest.TestCase):
    def test_validate_jwks_uri(self):
        # required
        metadata = OpenIDProviderMetadata()
        with self.assertRaises(ValueError) as cm:
            metadata.validate_jwks_uri()
        self.assertEqual('"jwks_uri" is required', str(cm.exception))

        metadata = OpenIDProviderMetadata({"jwks_uri": "http://authlib.org/jwks.json"})
        with self.assertRaises(ValueError) as cm:
            metadata.validate_jwks_uri()
        self.assertIn("https", str(cm.exception))

        metadata = OpenIDProviderMetadata({"jwks_uri": "https://authlib.org/jwks.json"})
        metadata.validate_jwks_uri()

    def test_validate_acr_values_supported(self):
        self._call_validate_array(
            "acr_values_supported", ["urn:mace:incommon:iap:silver"]
        )

    def test_validate_subject_types_supported(self):
        self._call_validate_array(
            "subject_types_supported", ["pairwise", "public"], required=True
        )
        self._call_contains_invalid_value("subject_types_supported", ["invalid"])

    def test_validate_id_token_signing_alg_values_supported(self):
        self._call_validate_array(
            "id_token_signing_alg_values_supported",
            ["RS256"],
            required=True,
        )
        metadata = OpenIDProviderMetadata(
            {"id_token_signing_alg_values_supported": ["none"]}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_id_token_signing_alg_values_supported()
        self.assertIn("RS256", str(cm.exception))

    def test_validate_id_token_encryption_alg_values_supported(self):
        self._call_validate_array(
            "id_token_encryption_alg_values_supported", ["A128KW"]
        )

    def test_validate_id_token_encryption_enc_values_supported(self):
        self._call_validate_array(
            "id_token_encryption_enc_values_supported", ["A128GCM"]
        )

    def test_validate_userinfo_signing_alg_values_supported(self):
        self._call_validate_array("userinfo_signing_alg_values_supported", ["RS256"])

    def test_validate_userinfo_encryption_alg_values_supported(self):
        self._call_validate_array(
            "userinfo_encryption_alg_values_supported", ["A128KW"]
        )

    def test_validate_userinfo_encryption_enc_values_supported(self):
        self._call_validate_array(
            "userinfo_encryption_enc_values_supported", ["A128GCM"]
        )

    def test_validate_request_object_signing_alg_values_supported(self):
        self._call_validate_array(
            "request_object_signing_alg_values_supported", ["none", "RS256"]
        )
        metadata = OpenIDProviderMetadata(
            {"request_object_signing_alg_values_supported": ["RS512"]}
        )
        with self.assertRaises(ValueError) as cm:
            metadata.validate_request_object_signing_alg_values_supported()
        self.assertIn("SHOULD support none and RS256", str(cm.exception))

    def test_validate_request_object_encryption_alg_values_supported(self):
        self._call_validate_array(
            "request_object_encryption_alg_values_supported", ["A128KW"]
        )

    def test_validate_request_object_encryption_enc_values_supported(self):
        self._call_validate_array(
            "request_object_encryption_enc_values_supported", ["A128GCM"]
        )

    def test_validate_display_values_supported(self):
        self._call_validate_array("display_values_supported", ["page", "touch"])
        self._call_contains_invalid_value("display_values_supported", ["invalid"])

    def test_validate_claim_types_supported(self):
        self._call_validate_array("claim_types_supported", ["normal"])
        self._call_contains_invalid_value("claim_types_supported", ["invalid"])
        metadata = OpenIDProviderMetadata()
        self.assertEqual(metadata.claim_types_supported, ["normal"])

    def test_validate_claims_supported(self):
        self._call_validate_array("claims_supported", ["sub"])

    def test_validate_claims_locales_supported(self):
        self._call_validate_array("claims_locales_supported", ["en-US"])

    def test_validate_claims_parameter_supported(self):
        self._call_validate_boolean("claims_parameter_supported")

    def test_validate_request_parameter_supported(self):
        self._call_validate_boolean("request_parameter_supported")

    def test_validate_request_uri_parameter_supported(self):
        self._call_validate_boolean("request_uri_parameter_supported", True)

    def test_validate_require_request_uri_registration(self):
        self._call_validate_boolean("require_request_uri_registration")

    def _call_validate_boolean(self, key, default_value=False):
        def _validate(metadata):
            getattr(metadata, "validate_" + key)()

        metadata = OpenIDProviderMetadata()
        _validate(metadata)
        self.assertEqual(getattr(metadata, key), default_value)

        metadata = OpenIDProviderMetadata({key: "str"})
        with self.assertRaises(ValueError) as cm:
            _validate(metadata)
        self.assertIn("MUST be boolean", str(cm.exception))
        metadata = OpenIDProviderMetadata({key: True})
        _validate(metadata)

    def _call_validate_array(self, key, valid_value, required=False):
        def _validate(metadata):
            getattr(metadata, "validate_" + key)()

        metadata = OpenIDProviderMetadata()
        if required:
            with self.assertRaises(ValueError) as cm:
                _validate(metadata)
            self.assertEqual(f'"{key}" is required', str(cm.exception))
        else:
            _validate(metadata)

        # not array
        metadata = OpenIDProviderMetadata({key: "foo"})
        with self.assertRaises(ValueError) as cm:
            _validate(metadata)
        self.assertIn("JSON array", str(cm.exception))

        # valid
        metadata = OpenIDProviderMetadata({key: valid_value})
        _validate(metadata)

    def _call_contains_invalid_value(self, key, invalid_value):
        metadata = OpenIDProviderMetadata({key: invalid_value})
        with self.assertRaises(ValueError) as cm:
            getattr(metadata, "validate_" + key)()
        self.assertEqual(f'"{key}" contains invalid values', str(cm.exception))



from flask import json

from authlib.integrations.sqla_oauth2 import create_revocation_endpoint

from .models import Client
from .models import Token
from .models import User
from .models import db
from .oauth2_server import TestCase
from .oauth2_server import create_authorization_server

RevocationEndpoint = create_revocation_endpoint(db.session, Token)


class RevokeTokenTest(TestCase):
    def prepare_data(self):
        app = self.app
        server = create_authorization_server(app)
        server.register_endpoint(RevocationEndpoint)

        @app.route("/oauth/revoke", methods=["POST"])
        def revoke_token():
            return server.create_endpoint_response("revocation")

        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="revoke-client",
            client_secret="revoke-secret",
        )
        client.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
            }
        )
        db.session.add(client)
        db.session.commit()

    def create_token(self):
        token = Token(
            user_id=1,
            client_id="revoke-client",
            token_type="bearer",
            access_token="a1",
            refresh_token="r1",
            scope="profile",
            expires_in=3600,
        )
        db.session.add(token)
        db.session.commit()

    def test_invalid_client(self):
        self.prepare_data()
        rv = self.client.post("/oauth/revoke")
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = {"Authorization": "invalid token_string"}
        rv = self.client.post("/oauth/revoke", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("invalid-client", "revoke-secret")
        rv = self.client.post("/oauth/revoke", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

        headers = self.create_basic_header("revoke-client", "invalid-secret")
        rv = self.client.post("/oauth/revoke", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_client")

    def test_invalid_token(self):
        self.prepare_data()
        headers = self.create_basic_header("revoke-client", "revoke-secret")
        rv = self.client.post("/oauth/revoke", headers=headers)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_request")

        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "invalid-token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)

        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "a1",
                "token_type_hint": "unsupported_token_type",
            },
            headers=headers,
        )
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "unsupported_token_type")

        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "a1",
                "token_type_hint": "refresh_token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)

    def test_revoke_token_with_hint(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("revoke-client", "revoke-secret")
        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "a1",
                "token_type_hint": "access_token",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)

    def test_revoke_token_without_hint(self):
        self.prepare_data()
        self.create_token()
        headers = self.create_basic_header("revoke-client", "revoke-secret")
        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "a1",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 200)

    def test_revoke_token_bound_to_client(self):
        self.prepare_data()
        self.create_token()

        client2 = Client(
            user_id=1,
            client_id="revoke-client-2",
            client_secret="revoke-secret-2",
        )
        client2.set_client_metadata(
            {
                "scope": "profile",
                "redirect_uris": ["http://localhost/authorized"],
            }
        )
        db.session.add(client2)
        db.session.commit()

        headers = self.create_basic_header("revoke-client-2", "revoke-secret-2")
        rv = self.client.post(
            "/oauth/revoke",
            data={
                "token": "a1",
            },
            headers=headers,
        )
        self.assertEqual(rv.status_code, 400)
        resp = json.loads(rv.data)
        self.assertEqual(resp["error"], "invalid_grant")



from tests.util import decode_response

from .oauth1_server import Client
from .oauth1_server import TestCase
from .oauth1_server import User
from .oauth1_server import create_authorization_server
from .oauth1_server import db


class AuthorizationWithCacheTest(TestCase):
    USE_CACHE = True

    def prepare_data(self):
        create_authorization_server(self.app, self.USE_CACHE, self.USE_CACHE)
        user = User(username="foo")
        db.session.add(user)
        db.session.commit()
        client = Client(
            user_id=user.id,
            client_id="client",
            client_secret="secret",
            default_redirect_uri="https://a.b",
        )
        db.session.add(client)
        db.session.commit()

    def test_invalid_authorization(self):
        self.prepare_data()
        url = "/oauth/authorize"

        # case 1
        rv = self.client.post(url, data={"user_id": "1"})
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "missing_required_parameter")
        self.assertIn("oauth_token", data["error_description"])

        # case 2
        rv = self.client.post(url, data={"user_id": "1", "oauth_token": "a"})
        data = decode_response(rv.data)
        self.assertEqual(data["error"], "invalid_token")

    def test_authorize_denied(self):
        self.prepare_data()
        initiate_url = "/oauth/initiate"
        authorize_url = "/oauth/authorize"

        rv = self.client.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "oob",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        rv = self.client.post(authorize_url, data={"oauth_token": data["oauth_token"]})
        self.assertEqual(rv.status_code, 302)
        self.assertIn("access_denied", rv.headers["Location"])
        self.assertIn("https://a.b", rv.headers["Location"])

        rv = self.client.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "https://i.test",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        rv = self.client.post(authorize_url, data={"oauth_token": data["oauth_token"]})
        self.assertEqual(rv.status_code, 302)
        self.assertIn("access_denied", rv.headers["Location"])
        self.assertIn("https://i.test", rv.headers["Location"])

    def test_authorize_granted(self):
        self.prepare_data()
        initiate_url = "/oauth/initiate"
        authorize_url = "/oauth/authorize"

        rv = self.client.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "oob",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        rv = self.client.post(
            authorize_url, data={"user_id": "1", "oauth_token": data["oauth_token"]}
        )
        self.assertEqual(rv.status_code, 302)
        self.assertIn("oauth_verifier", rv.headers["Location"])
        self.assertIn("https://a.b", rv.headers["Location"])

        rv = self.client.post(
            initiate_url,
            data={
                "oauth_consumer_key": "client",
                "oauth_callback": "https://i.test",
                "oauth_signature_method": "PLAINTEXT",
                "oauth_signature": "secret&",
            },
        )
        data = decode_response(rv.data)
        self.assertIn("oauth_token", data)

        rv = self.client.post(
            authorize_url, data={"user_id": "1", "oauth_token": data["oauth_token"]}
        )
        self.assertEqual(rv.status_code, 302)
        self.assertIn("oauth_verifier", rv.headers["Location"])
        self.assertIn("https://i.test", rv.headers["Location"])


class AuthorizationNoCacheTest(AuthorizationWithCacheTest):
    USE_CACHE = False



import unittest

from authlib.oauth2.rfc7662 import IntrospectionToken


class IntrospectionTokenTest(unittest.TestCase):
    def test_client_id(self):
        token = IntrospectionToken()
        self.assertIsNone(token.client_id)
        self.assertIsNone(token.get_client_id())

        token = IntrospectionToken({"client_id": "foo"})
        self.assertEqual(token.client_id, "foo")
        self.assertEqual(token.get_client_id(), "foo")

    def test_scope(self):
        token = IntrospectionToken()
        self.assertIsNone(token.scope)
        self.assertIsNone(token.get_scope())

        token = IntrospectionToken({"scope": "foo"})
        self.assertEqual(token.scope, "foo")
        self.assertEqual(token.get_scope(), "foo")

    def test_expires_in(self):
        token = IntrospectionToken()
        self.assertEqual(token.get_expires_in(), 0)

    def test_expires_at(self):
        token = IntrospectionToken()
        self.assertIsNone(token.exp)
        self.assertEqual(token.get_expires_at(), 0)

        token = IntrospectionToken({"exp": 3600})
        self.assertEqual(token.exp, 3600)
        self.assertEqual(token.get_expires_at(), 3600)

    def test_all_attributes(self):
        # https://tools.ietf.org/html/rfc7662#section-2.2
        token = IntrospectionToken()
        self.assertIsNone(token.active)
        self.assertIsNone(token.scope)
        self.assertIsNone(token.client_id)
        self.assertIsNone(token.username)
        self.assertIsNone(token.token_type)
        self.assertIsNone(token.exp)
        self.assertIsNone(token.iat)
        self.assertIsNone(token.nbf)
        self.assertIsNone(token.sub)
        self.assertIsNone(token.aud)
        self.assertIsNone(token.iss)
        self.assertIsNone(token.jti)

    def test_invalid_attr(self):
        token = IntrospectionToken()
        self.assertRaises(AttributeError, lambda: token.invalid)



from io import StringIO
from unittest import TestCase
from unittest import mock

import requests

from authlib.common.encoding import to_unicode
from authlib.integrations.requests_client import OAuth1Session
from authlib.integrations.requests_client import OAuthError
from authlib.oauth1 import SIGNATURE_PLAINTEXT
from authlib.oauth1 import SIGNATURE_RSA_SHA1
from authlib.oauth1 import SIGNATURE_TYPE_BODY
from authlib.oauth1 import SIGNATURE_TYPE_QUERY
from authlib.oauth1.rfc5849.util import escape

from ..util import mock_text_response
from ..util import read_key_file

TEST_RSA_OAUTH_SIGNATURE = (
    "j8WF8PGjojT82aUDd2EL%2Bz7HCoHInFzWUpiEKMCy%2BJ2cYHWcBS7mXlmFDLgAKV0"
    "P%2FyX4TrpXODYnJ6dRWdfghqwDpi%2FlQmB2jxCiGMdJoYxh3c5zDf26gEbGdP6D7O"
    "Ssp5HUnzH6sNkmVjuE%2FxoJcHJdc23H6GhOs7VJ2LWNdbhKWP%2FMMlTrcoQDn8lz"
    "%2Fb24WsJ6ae1txkUzpFOOlLM8aTdNtGL4OtsubOlRhNqnAFq93FyhXg0KjzUyIZzmMX"
    "9Vx90jTks5QeBGYcLE0Op2iHb2u%2FO%2BEgdwFchgEwE5LgMUyHUI4F3Wglp28yHOAM"
    "jPkI%2FkWMvpxtMrU3Z3KN31WQ%3D%3D"
)


class OAuth1SessionTest(TestCase):
    def test_no_client_id(self):
        self.assertRaises(ValueError, lambda: OAuth1Session(None))

    def test_signature_types(self):
        def verify_signature(getter):
            def fake_send(r, **kwargs):
                signature = to_unicode(getter(r))
                self.assertIn("oauth_signature", signature)
                resp = mock.MagicMock(spec=requests.Response)
                resp.cookies = []
                return resp

            return fake_send

        header = OAuth1Session("foo")
        header.send = verify_signature(lambda r: r.headers["Authorization"])
        header.post("https://i.b")

        query = OAuth1Session("foo", signature_type=SIGNATURE_TYPE_QUERY)
        query.send = verify_signature(lambda r: r.url)
        query.post("https://i.b")

        body = OAuth1Session("foo", signature_type=SIGNATURE_TYPE_BODY)
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body.send = verify_signature(lambda r: r.body)
        body.post("https://i.b", headers=headers, data="")

    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
    def test_signature_methods(self, generate_nonce, generate_timestamp):
        generate_nonce.return_value = "abc"
        generate_timestamp.return_value = "123"

        signature = ", ".join(
            [
                'OAuth oauth_nonce="abc"',
                'oauth_timestamp="123"',
                'oauth_version="1.0"',
                'oauth_signature_method="HMAC-SHA1"',
                'oauth_consumer_key="foo"',
                'oauth_signature="h2sRqLArjhlc5p3FTkuNogVHlKE%3D"',
            ]
        )
        auth = OAuth1Session("foo")
        auth.send = self.verify_signature(signature)
        auth.post("https://i.b")

        signature = (
            "OAuth "
            'oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
            'oauth_signature_method="PLAINTEXT", oauth_consumer_key="foo", '
            'oauth_signature="%26"'
        )
        auth = OAuth1Session("foo", signature_method=SIGNATURE_PLAINTEXT)
        auth.send = self.verify_signature(signature)
        auth.post("https://i.b")

        signature = (
            "OAuth "
            'oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
            'oauth_signature_method="RSA-SHA1", oauth_consumer_key="foo", '
            f'oauth_signature="{TEST_RSA_OAUTH_SIGNATURE}"'
        )

        rsa_key = read_key_file("rsa_private.pem")
        auth = OAuth1Session(
            "foo", signature_method=SIGNATURE_RSA_SHA1, rsa_key=rsa_key
        )
        auth.send = self.verify_signature(signature)
        auth.post("https://i.b")

    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
    def test_binary_upload(self, generate_nonce, generate_timestamp):
        generate_nonce.return_value = "abc"
        generate_timestamp.return_value = "123"
        fake_xml = StringIO("hello world")
        headers = {"Content-Type": "application/xml"}

        def fake_send(r, **kwargs):
            auth_header = r.headers["Authorization"]
            self.assertIn("oauth_body_hash", auth_header)

        auth = OAuth1Session("foo", force_include_body=True)
        auth.send = fake_send
        auth.post("https://i.b", headers=headers, files=[("fake", fake_xml)])

    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
    @mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
    def test_nonascii(self, generate_nonce, generate_timestamp):
        generate_nonce.return_value = "abc"
        generate_timestamp.return_value = "123"
        signature = (
            'OAuth oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
            'oauth_signature_method="HMAC-SHA1", oauth_consumer_key="foo", '
            'oauth_signature="W0haoue5IZAZoaJiYCtfqwMf8x8%3D"'
        )
        auth = OAuth1Session("foo")
        auth.send = self.verify_signature(signature)
        auth.post("https://i.b?cjk=%E5%95%A6%E5%95%A6")

    def test_redirect_uri(self):
        sess = OAuth1Session("foo")
        self.assertIsNone(sess.redirect_uri)
        url = "https://i.b"
        sess.redirect_uri = url
        self.assertEqual(sess.redirect_uri, url)

    def test_set_token(self):
        sess = OAuth1Session("foo")
        try:
            sess.token = {}
        except OAuthError as exc:
            self.assertEqual(exc.error, "missing_token")

        sess.token = {"oauth_token": "a", "oauth_token_secret": "b"}
        self.assertIsNone(sess.token["oauth_verifier"])
        sess.token = {"oauth_token": "a", "oauth_verifier": "c"}
        self.assertEqual(sess.token["oauth_token_secret"], "b")
        self.assertEqual(sess.token["oauth_verifier"], "c")

        sess.token = None
        self.assertIsNone(sess.token["oauth_token"])
        self.assertIsNone(sess.token["oauth_token_secret"])
        self.assertIsNone(sess.token["oauth_verifier"])

    def test_create_authorization_url(self):
        auth = OAuth1Session("foo")
        url = "https://example.comm/authorize"
        token = "asluif023sf"
        auth_url = auth.create_authorization_url(url, request_token=token)
        self.assertEqual(auth_url, url + "?oauth_token=" + token)
        redirect_uri = "https://c.b"
        auth = OAuth1Session("foo", redirect_uri=redirect_uri)
        auth_url = auth.create_authorization_url(url, request_token=token)
        self.assertIn(escape(redirect_uri), auth_url)

    def test_parse_response_url(self):
        url = "https://i.b/callback?oauth_token=foo&oauth_verifier=bar"
        auth = OAuth1Session("foo")
        resp = auth.parse_authorization_response(url)
        self.assertEqual(resp["oauth_token"], "foo")
        self.assertEqual(resp["oauth_verifier"], "bar")
        for k, v in resp.items():
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, str))

    def test_fetch_request_token(self):
        auth = OAuth1Session("foo", realm="A")
        auth.send = mock_text_response("oauth_token=foo")
        resp = auth.fetch_request_token("https://example.com/token")
        self.assertEqual(resp["oauth_token"], "foo")
        for k, v in resp.items():
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, str))

        resp = auth.fetch_request_token("https://example.com/token")
        self.assertEqual(resp["oauth_token"], "foo")

    def test_fetch_request_token_with_optional_arguments(self):
        auth = OAuth1Session("foo")
        auth.send = mock_text_response("oauth_token=foo")
        resp = auth.fetch_request_token(
            "https://example.com/token", verify=False, stream=True
        )
        self.assertEqual(resp["oauth_token"], "foo")
        for k, v in resp.items():
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, str))

    def test_fetch_access_token(self):
        auth = OAuth1Session("foo", verifier="bar")
        auth.send = mock_text_response("oauth_token=foo")
        resp = auth.fetch_access_token("https://example.com/token")
        self.assertEqual(resp["oauth_token"], "foo")
        for k, v in resp.items():
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, str))

        auth = OAuth1Session("foo", verifier="bar")
        auth.send = mock_text_response('{"oauth_token":"foo"}')
        resp = auth.fetch_access_token("https://example.com/token")
        self.assertEqual(resp["oauth_token"], "foo")

        auth = OAuth1Session("foo")
        auth.send = mock_text_response("oauth_token=foo")
        resp = auth.fetch_access_token("https://example.com/token", verifier="bar")
        self.assertEqual(resp["oauth_token"], "foo")

    def test_fetch_access_token_with_optional_arguments(self):
        auth = OAuth1Session("foo", verifier="bar")
        auth.send = mock_text_response("oauth_token=foo")
        resp = auth.fetch_access_token(
            "https://example.com/token", verify=False, stream=True
        )
        self.assertEqual(resp["oauth_token"], "foo")
        for k, v in resp.items():
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, str))

    def _test_fetch_access_token_raises_error(self, session):
        """Assert that an error is being raised whenever there's no verifier
        passed in to the client.
        """
        session.send = mock_text_response("oauth_token=foo")

        # Use a try-except block so that we can assert on the exception message
        # being raised and also keep the Python2.6 compatibility where
        # assertRaises is not a context manager.
        try:
            session.fetch_access_token("https://example.com/token")
        except OAuthError as exc:
            self.assertEqual(exc.error, "missing_verifier")

    def test_fetch_token_invalid_response(self):
        auth = OAuth1Session("foo")
        auth.send = mock_text_response("not valid urlencoded response!")
        self.assertRaises(
            ValueError, auth.fetch_request_token, "https://example.com/token"
        )

        for code in (400, 401, 403):
            auth.send = mock_text_response("valid=response", code)
            # use try/catch rather than self.assertRaises, so we can
            # assert on the properties of the exception
            try:
                auth.fetch_request_token("https://example.com/token")
            except OAuthError as err:
                self.assertEqual(err.error, "fetch_token_denied")
            else:  # no exception raised
                self.fail("ValueError not raised")

    def test_fetch_access_token_missing_verifier(self):
        self._test_fetch_access_token_raises_error(OAuth1Session("foo"))

    def test_fetch_access_token_has_verifier_is_none(self):
        session = OAuth1Session("foo")
        session.auth.verifier = None
        self._test_fetch_access_token_raises_error(session)

    def verify_signature(self, signature):
        def fake_send(r, **kwargs):
            auth_header = to_unicode(r.headers["Authorization"])
            self.assertEqual(auth_header, signature)
            resp = mock.MagicMock(spec=requests.Response)
            resp.cookies = []
            return resp

        return fake_send



import base64
import unittest

from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import parameters
from authlib.oauth2.rfc6749 import util


class OAuth2ParametersTest(unittest.TestCase):
    def test_parse_authorization_code_response(self):
        self.assertRaises(
            errors.MissingCodeException,
            parameters.parse_authorization_code_response,
            "https://i.b/?state=c",
        )

        self.assertRaises(
            errors.MismatchingStateException,
            parameters.parse_authorization_code_response,
            "https://i.b/?code=a&state=c",
            "b",
        )

        url = "https://i.b/?code=a&state=c"
        rv = parameters.parse_authorization_code_response(url, "c")
        self.assertEqual(rv, {"code": "a", "state": "c"})

    def test_parse_implicit_response(self):
        self.assertRaises(
            errors.MissingTokenException,
            parameters.parse_implicit_response,
            "https://i.b/#a=b",
        )

        self.assertRaises(
            errors.MissingTokenTypeException,
            parameters.parse_implicit_response,
            "https://i.b/#access_token=a",
        )

        self.assertRaises(
            errors.MismatchingStateException,
            parameters.parse_implicit_response,
            "https://i.b/#access_token=a&token_type=bearer&state=c",
            "abc",
        )

        url = "https://i.b/#access_token=a&token_type=bearer&state=c"
        rv = parameters.parse_implicit_response(url, "c")
        self.assertEqual(
            rv, {"access_token": "a", "token_type": "bearer", "state": "c"}
        )

    def test_prepare_grant_uri(self):
        grant_uri = parameters.prepare_grant_uri(
            "https://i.b/authorize", "dev", "code", max_age=0
        )
        self.assertEqual(
            grant_uri,
            "https://i.b/authorize?response_type=code&client_id=dev&max_age=0",
        )


class OAuth2UtilTest(unittest.TestCase):
    def test_list_to_scope(self):
        self.assertEqual(util.list_to_scope(["a", "b"]), "a b")
        self.assertEqual(util.list_to_scope("a b"), "a b")
        self.assertIsNone(util.list_to_scope(None))

    def test_scope_to_list(self):
        self.assertEqual(util.scope_to_list("a b"), ["a", "b"])
        self.assertEqual(util.scope_to_list(["a", "b"]), ["a", "b"])
        self.assertIsNone(util.scope_to_list(None))

    def test_extract_basic_authorization(self):
        self.assertEqual(util.extract_basic_authorization({}), (None, None))
        self.assertEqual(
            util.extract_basic_authorization({"Authorization": "invalid"}), (None, None)
        )

        text = "Basic invalid-base64"
        self.assertEqual(
            util.extract_basic_authorization({"Authorization": text}), (None, None)
        )

        text = "Basic {}".format(base64.b64encode(b"a").decode())
        self.assertEqual(
            util.extract_basic_authorization({"Authorization": text}), ("a", None)
        )

        text = "Basic {}".format(base64.b64encode(b"a:b").decode())
        self.assertEqual(
            util.extract_basic_authorization({"Authorization": text}), ("a", "b")
        )



import json

from django.http import JsonResponse

from authlib.integrations.django_oauth2 import BearerTokenValidator
from authlib.integrations.django_oauth2 import ResourceProtector

from .models import Client
from .models import OAuth2Token
from .models import User
from .oauth2_server import TestCase

require_oauth = ResourceProtector()
require_oauth.register_token_validator(BearerTokenValidator(OAuth2Token))


class ResourceProtectorTest(TestCase):
    def prepare_data(self, expires_in=3600, scope="profile"):
        user = User(username="foo")
        user.save()
        client = Client(
            user_id=user.pk,
            client_id="client",
            client_secret="secret",
            scope="profile",
        )
        client.save()

        token = OAuth2Token(
            user_id=user.pk,
            client_id=client.client_id,
            token_type="bearer",
            access_token="a1",
            scope=scope,
            expires_in=expires_in,
        )
        token.save()

    def test_invalid_token(self):
        @require_oauth("profile")
        def get_user_profile(request):
            user = request.oauth_token.user
            return JsonResponse(dict(sub=user.pk, username=user.username))

        self.prepare_data()

        request = self.factory.get("/user")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "missing_authorization")

        request = self.factory.get("/user", HTTP_AUTHORIZATION="invalid token")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "unsupported_token_type")

        request = self.factory.get("/user", HTTP_AUTHORIZATION="bearer token")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_token")

    def test_expired_token(self):
        self.prepare_data(-10)

        @require_oauth("profile")
        def get_user_profile(request):
            user = request.oauth_token.user
            return JsonResponse(dict(sub=user.pk, username=user.username))

        request = self.factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "invalid_token")

    def test_insufficient_token(self):
        self.prepare_data()

        @require_oauth("email")
        def get_user_email(request):
            user = request.oauth_token.user
            return JsonResponse(dict(email=user.email))

        request = self.factory.get("/user/email", HTTP_AUTHORIZATION="bearer a1")
        resp = get_user_email(request)
        self.assertEqual(resp.status_code, 403)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "insufficient_scope")

    def test_access_resource(self):
        self.prepare_data()

        @require_oauth("profile", optional=True)
        def get_user_profile(request):
            if request.oauth_token:
                user = request.oauth_token.user
                return JsonResponse(dict(sub=user.pk, username=user.username))
            return JsonResponse(dict(sub=0, username="anonymous"))

        request = self.factory.get("/user")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertEqual(data["username"], "anonymous")

        request = self.factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
        resp = get_user_profile(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertEqual(data["username"], "foo")

    def test_scope_operator(self):
        self.prepare_data()

        @require_oauth(["profile email"])
        def operator_and(request):
            user = request.oauth_token.user
            return JsonResponse(dict(sub=user.pk, username=user.username))

        @require_oauth(["profile", "email"])
        def operator_or(request):
            user = request.oauth_token.user
            return JsonResponse(dict(sub=user.pk, username=user.username))

        request = self.factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
        resp = operator_and(request)
        self.assertEqual(resp.status_code, 403)
        data = json.loads(resp.content)
        self.assertEqual(data["error"], "insufficient_scope")

        resp = operator_or(request)
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.content)
        self.assertEqual(data["username"], "foo")



import unittest

from authlib.jose.errors import InvalidClaimError
from authlib.jose.errors import MissingClaimError
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core import HybridIDToken
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core import UserInfo
from authlib.oidc.core import get_claim_cls_by_response_type


class IDTokenTest(unittest.TestCase):
    def test_essential_claims(self):
        claims = CodeIDToken({}, {})
        self.assertRaises(MissingClaimError, claims.validate)
        claims = CodeIDToken(
            {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
        )
        claims.validate(1000)

    def test_validate_auth_time(self):
        claims = CodeIDToken(
            {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
        )
        claims.params = {"max_age": 100}
        self.assertRaises(MissingClaimError, claims.validate, 1000)

        claims["auth_time"] = "foo"
        self.assertRaises(InvalidClaimError, claims.validate, 1000)

    def test_validate_nonce(self):
        claims = CodeIDToken(
            {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
        )
        claims.params = {"nonce": "foo"}
        self.assertRaises(MissingClaimError, claims.validate, 1000)
        claims["nonce"] = "bar"
        self.assertRaises(InvalidClaimError, claims.validate, 1000)
        claims["nonce"] = "foo"
        claims.validate(1000)

    def test_validate_amr(self):
        claims = CodeIDToken(
            {
                "iss": "1",
                "sub": "1",
                "aud": "1",
                "exp": 10000,
                "iat": 100,
                "amr": "invalid",
            },
            {},
        )
        self.assertRaises(InvalidClaimError, claims.validate, 1000)

    def test_validate_azp(self):
        claims = CodeIDToken(
            {
                "iss": "1",
                "sub": "1",
                "aud": "1",
                "exp": 10000,
                "iat": 100,
            },
            {},
        )
        claims.params = {"client_id": "2"}
        self.assertRaises(MissingClaimError, claims.validate, 1000)

        claims["azp"] = "1"
        self.assertRaises(InvalidClaimError, claims.validate, 1000)

        claims["azp"] = "2"
        claims.validate(1000)

    def test_validate_at_hash(self):
        claims = CodeIDToken(
            {
                "iss": "1",
                "sub": "1",
                "aud": "1",
                "exp": 10000,
                "iat": 100,
                "at_hash": "a",
            },
            {},
        )
        claims.params = {"access_token": "a"}

        # invalid alg won't raise
        claims.header = {"alg": "HS222"}
        claims.validate(1000)

        claims.header = {"alg": "HS256"}
        self.assertRaises(InvalidClaimError, claims.validate, 1000)

    def test_implicit_id_token(self):
        claims = ImplicitIDToken(
            {
                "iss": "1",
                "sub": "1",
                "aud": "1",
                "exp": 10000,
                "iat": 100,
                "nonce": "a",
            },
            {},
        )
        claims.params = {"access_token": "a"}
        self.assertRaises(MissingClaimError, claims.validate, 1000)

    def test_hybrid_id_token(self):
        claims = HybridIDToken(
            {
                "iss": "1",
                "sub": "1",
                "aud": "1",
                "exp": 10000,
                "iat": 100,
                "nonce": "a",
            },
            {},
        )
        claims.validate(1000)

        claims.params = {"code": "a"}
        self.assertRaises(MissingClaimError, claims.validate, 1000)

        # invalid alg won't raise
        claims.header = {"alg": "HS222"}
        claims["c_hash"] = "a"
        claims.validate(1000)

        claims.header = {"alg": "HS256"}
        self.assertRaises(InvalidClaimError, claims.validate, 1000)

    def test_get_claim_cls_by_response_type(self):
        cls = get_claim_cls_by_response_type("id_token")
        self.assertEqual(cls, ImplicitIDToken)
        cls = get_claim_cls_by_response_type("code")
        self.assertEqual(cls, CodeIDToken)
        cls = get_claim_cls_by_response_type("code id_token")
        self.assertEqual(cls, HybridIDToken)
        cls = get_claim_cls_by_response_type("none")
        self.assertIsNone(cls)


class UserInfoTest(unittest.TestCase):
    def test_getattribute(self):
        user = UserInfo({"sub": "1"})
        self.assertEqual(user.sub, "1")
        self.assertIsNone(user.email, None)
        self.assertRaises(AttributeError, lambda: user.invalid)
